#ifndef _EFL_SELECTION_MANAGER_EO_H_
#define _EFL_SELECTION_MANAGER_EO_H_

#ifndef _EFL_SELECTION_MANAGER_EO_CLASS_TYPE
#define _EFL_SELECTION_MANAGER_EO_CLASS_TYPE

typedef Eo Efl_Selection_Manager;

#endif

#ifndef _EFL_SELECTION_MANAGER_EO_TYPES
#define _EFL_SELECTION_MANAGER_EO_TYPES


#endif
#define EFL_SELECTION_MANAGER_CLASS efl_selection_manager_class_get()

EWAPI const Efl_Class *efl_selection_manager_class_get(void);

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief Set selection
 *
 * @param[in] obj The object.
 * @param[in] owner Seleciton owner
 * @param[in] type Selection type
 * @param[in] format Selection format
 * @param[in] data Selection data
 * @param[in] seat Specified seat for multiple seats case.
 *
 * @return Future for tracking when the selection is lost
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI Eina_Future *efl_selection_manager_selection_set(Eo *obj, Efl_Object *owner, Efl_Selection_Type type, Efl_Selection_Format format, Eina_Slice data, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief Get selection
 *
 * @param[in] obj The object.
 * @param[in] request Seleciton owner
 * @param[in] type Selection type
 * @param[in] format Selection Format
 * @param[in] data_func Data ready function pointer
 * @param[in] seat Specified seat for multiple seats case.
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI void efl_selection_manager_selection_get(Eo *obj, const Efl_Object *request, Efl_Selection_Type type, Efl_Selection_Format format, void *data_func_data, Efl_Selection_Data_Ready data_func, Eina_Free_Cb data_func_free_cb, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
EOAPI void efl_selection_manager_selection_clear(Eo *obj, Efl_Object *owner, Efl_Selection_Type type, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief Check if the request object has selection or not
 *
 * @param[in] obj The object.
 * @param[in] request Request object
 * @param[in] type Selection type
 * @param[in] seat Specified seat for multiple seats case.
 *
 * @return EINA_TRUE if the request object has selection, otherwise, EINA_FALSE
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI Eina_Bool efl_selection_manager_selection_has_owner(Eo *obj, Efl_Object *request, Efl_Selection_Type type, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief This starts a drag and drop process at the drag side. During
 * dragging, there are three events emitted as belows: -
 * EFL_UI_DND_EVENT_DRAG_POS - EFL_UI_DND_EVENT_DRAG_ACCEPT -
 * EFL_UI_DND_EVENT_DRAG_DONE
 *
 * @param[in] obj The object.
 * @param[in] drag_obj Drag object
 * @param[in] format Data format
 * @param[in] data Data to transfer
 * @param[in] action Action when data is transferred
 * @param[in] icon_func Function pointer to create icon
 * @param[in] seat Specified seat for multiple seats case.
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI void efl_selection_manager_drag_start(Eo *obj, Efl_Object *drag_obj, Efl_Selection_Format format, Eina_Slice data, Efl_Selection_Action action, void *icon_func_data, Efl_Dnd_Drag_Icon_Create icon_func, Eina_Free_Cb icon_func_free_cb, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief This sets the action for the drag
 *
 * @param[in] obj The object.
 * @param[in] drag_obj Drag object
 * @param[in] action Drag action
 * @param[in] seat Specified seat for multiple seats case.
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI void efl_selection_manager_drag_action_set(Eo *obj, Efl_Object *drag_obj, Efl_Selection_Action action, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief This cancels the on-going drag
 *
 * @param[in] obj The object.
 * @param[in] drag_obj Drag object
 * @param[in] seat Specified seat for multiple seats case.
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI void efl_selection_manager_drag_cancel(Eo *obj, Efl_Object *drag_obj, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief This registers a drag for items in a container. Many items can be
 * dragged at a time. During dragging, there are three events emitted: -
 * EFL_UI_DND_EVENT_DRAG_POS - EFL_UI_DND_EVENT_DRAG_ACCEPT -
 * EFL_UI_DND_EVENT_DRAG_DONE.
 *
 * @param[in] obj The object.
 * @param[in] cont Container object
 * @param[in] time_to_drag Time since mouse down happens to drag starts
 * @param[in] anim_duration animation duration
 * @param[in] data_func Data and its format
 * @param[in] item_func Item to determine drag start
 * @param[in] icon_func Icon used during drag
 * @param[in] icon_list_func Icons used for animations
 * @param[in] seat Specified seat for multiple seats case
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI void efl_selection_manager_container_drag_item_add(Eo *obj, Efl_Object *cont, double time_to_drag, double anim_duration, void *data_func_data, Efl_Dnd_Drag_Data_Get data_func, Eina_Free_Cb data_func_free_cb, void *item_func_data, Efl_Dnd_Item_Get item_func, Eina_Free_Cb item_func_free_cb, void *icon_func_data, Efl_Dnd_Drag_Icon_Create icon_func, Eina_Free_Cb icon_func_free_cb, void *icon_list_func_data, Efl_Dnd_Drag_Icon_List_Create icon_list_func, Eina_Free_Cb icon_list_func_free_cb, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief Remove drag function of items in the container object.
 *
 * @param[in] obj The object.
 * @param[in] cont Container object
 * @param[in] seat Specified seat for multiple seats case
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI void efl_selection_manager_container_drag_item_del(Eo *obj, Efl_Object *cont, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief Add a dropable target. There are four events emitted: -
 * EFL_UI_DND_DROP_DRAG_ENTER - EFL_UI_DND_DROP_DRAG_LEAVE -
 * EFL_UI_DND_DROP_DRAG_POS - EFL_UI_DND_DROP_DRAG_DROP.
 *
 * @param[in] obj The object.
 * @param[in] target_obj Drop target
 * @param[in] format Accepted data format
 * @param[in] seat Specified seat for multiple seats case.
 *
 * @return @c true on success, @c false otherwise
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI Eina_Bool efl_selection_manager_drop_target_add(Eo *obj, Efl_Object *target_obj, Efl_Selection_Format format, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief Remove a dropable target
 *
 * @param[in] obj The object.
 * @param[in] target_obj Drop target
 * @param[in] format Accepted data format
 * @param[in] seat Specified seat for multiple seats case.
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI void efl_selection_manager_drop_target_del(Eo *obj, Efl_Object *target_obj, Efl_Selection_Format format, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief Add dropable target for a container in which items can drop to it
 *
 * @param[in] obj The object.
 * @param[in] cont Container object
 * @param[in] format Accepted data formats
 * @param[in] item_func Get item at specific position
 * @param[in] seat Specified seat for multiple seats case.
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI void efl_selection_manager_container_drop_item_add(Eo *obj, Efl_Object *cont, Efl_Selection_Format format, void *item_func_data, Efl_Dnd_Item_Get item_func, Eina_Free_Cb item_func_free_cb, unsigned int seat);
#endif

#ifdef EFL_SELECTION_MANAGER_BETA
/**
 * @brief Remove dropable target for the container
 *
 * @param[in] obj The object.
 * @param[in] cont Container object
 * @param[in] seat Specified seat for multiple seats case.
 *
 * @ingroup Efl_Selection_Manager
 */
EOAPI void efl_selection_manager_container_drop_item_del(Eo *obj, Efl_Object *cont, unsigned int seat);
#endif

#endif
